// opencommandio/core.js
// `````````````````````

var http = require('http');
var fs = require('fs');
var qs = require('querystring');
var io = require('socket.io');
var data = require('./servers.js');
var ssh = require('ssh2');

var count = 0;

console.log(data);

var app = http.createServer(webServer);
app.listen('8081', '127.0.0.1');

io.listen(app).sockets.on('connection', function(socket) {
	
	// notice classes: success|info|warning|danger
	socket.emit('notice', ['success', data.motd]);
	
	socket.on('run', function(detail) {
		
		if (! detail.hosts.length) {
			socket.emit('notice', ['warning', 'No servers selected.']);
		}
		
		for (var i = 0, e = detail.hosts.length; i !== e; ++i) {
			if ( ! data.servers.hasOwnProperty( detail.hosts[i] )) {
				socket.emit('notice', ['danger', 'Unknown or removed server "'+detail.hosts[i]+'" in request']);
				return;
			}
		}
		
		socket.emit('notice', ['info', 'Starting commands on '+e+' server(s)...']);
		for (var i = 0; i !== e; ++i) {
			handleExec(socket, detail.hosts[i], detail.src);
		}
	});
	
});

function memlen(str) {
	return Buffer.byteLength(str, 'utf8');
}

function handleExec(socket, serverKey, script) {
	
	var id = count;
	count += Math.floor(Math.random() * 4)+1;
	
	var c = new ssh();
	c.on('connect', function() {
		socket.emit('ssh', {'id': id, 'type':'info', 'message': 'SSH connected, logging in...'});
	});
	
	c.on('ready', function() {
		socket.emit('ssh', {'id': id, 'type':'info', 'message': 'Logged in, sending script...'});
		
		c.exec(
			'F=`mktemp`; head -c '+memlen(script)+' - > $F ; chmod +x $F ; $F ; echo Return value: $? ; rm $F ',
			function(err, stream) {
				
				stream.on('data', function(chunk, extended) {
					if (extended === 'stderr') {
						socket.emit('ssh', {'id': id, 'type':'stderr', 'message':chunk.toString('utf8')});
					} else {
						socket.emit('ssh', {'id': id, 'type':'stdout', 'message':chunk.toString('utf8')});
					}
					console.log(chunk.toString('utf8'));
				});
				
				var done_called = false;
				var done = function() {
					if (done_called) return;
					
					done_called = true;
					c.end();
				};
				
				stream.on('end', done);				
				stream.on('exit', done);
				stream.on('close', done);
				
				stream.write(script, 'UTF-8');
			}
		);
		
	});
	c.on('error', function() {
		socket.emit('ssh', {'id': id, 'type':'info', 'message': 'SSH connection error.'});
	});
	
	var closed_called = false;
	var closed = function() {
		if (closed_called) return;
		closed_called = true;
		
		socket.emit('ssh', {'id': id, 'type':'info', 'message': 'SSH connection closed.'});
	}
	
	c.on('end', closed);
	c.on('close', closed);
	
	socket.emit('ssh', {'id': id, 'type':'info', 'message': 'Connecting to server "'+serverKey+'"...'});
	c.connect( data.servers[serverKey] );
}

/**
 * n.b. strips leading forwardslash too
 * 
 * @param {String} sz
 * @returns {String}
 */
function escfile(sz) {
	return (''+sz).replace(/[^a-z\.]/g, '');
}

function redirect(res, url) {
	res.writeHead(302, {'Location': url});
	res.end();
}

function notfound(res) {
	res.writeHead(404, {'Content-Type': 'text/html'});
	res.end('<h1>404 not found</h1>');
}

/**
 * Retrieve a MIME type for a given filename/path. Defaults to text/plain for
 *   unknown extensions.
 * 
 * @param {String} name
 * @returns {String}
 */
function mimetype(name) {
	var extension = name.replace(/.+\.([^\.]+)/, '$1');
	switch(extension) {
		case 'css':
			return 'text/css';
		case 'js':
			return 'text/javascript';
		case 'htm':
		case 'html':
			return 'text/html';
		case 'ico':
			return 'image/x-icon';
		default:
			return 'text/plain';
	}
}

function handlePost(body, req, res) {	
	switch(req.url) {
		// no post actions are registered
		default: {
			notfound(res);
		} break;
	}		
}

function handleGet(req, res) {
	switch(req.url) {
		
		// Send back configuration
		case '/oci/config.js': {
			res.writeHead(200, {'Content-Type': 'text/javascript'});
			
			// Censor connection details
			var datasafe = JSON.parse(JSON.stringify(data)); // inefficient deep copy
			for (var i in datasafe.servers) {
				if (! datasafe.servers.hasOwnProperty(i)) continue; // js needs real hashmaps
				datasafe.servers[i] = true;
			}
			
			// Send to client
			res.end("oci_data = "+JSON.stringify(datasafe)+";"); // TODO remove all passwords + ssh keys
		} break;
		
		// Static file serving
		default: {
			var requrl = req.url || 'index.htm';
			var staticpath = __dirname + "/static/" + escfile(requrl); // strips leading /
			res.writeHead(200, {'Content-Type': mimetype(staticpath)});
			fs.readFile(staticpath, function(err, data) {
				if (!err) {
					res.end(data);
				} else notfound(res);
			});
		} break;
	}
}

function webServer(req, res) {
	if (req.method === 'POST') {
		var body = '';
		req.on('data', function(data) {
			body += data;
			if (body.length > 1e6 /* 1MB post size limit */) {
				req.connection.destroy(); // die
				body = '';
			}
		});
		req.on('end', function() {
			handlePost(qs.parse(body), req, res);
		});
		
	} else if (req.method === 'GET') {
		return handleGet(req, res);
		
	} else {
		res.writeHead(405); // Method not allowed
		res.end();
	}
}
